/******************************************************************************
 *
 *    Name: UIDialogLibWebKit.c
 *
 *    $Id: //LinuxBasedReceivers/main/ClientEngine/unix/UIDialogLib/WebKit/UIDialogLibWebKit.c#1 $
 *
 *    Purpose: A WebKit implementation for the UI Dialog Library.
 *
 *    Copyright 2017 Citrix Systems, Inc. All rights reserved.
 *
 *****************************************************************************/


#include "UIDialogLibWebKit.h"
#include "LinuxCommonUtilsSDK.cpp"

__attribute__((visibility("hidden"))) extern trace g_t1;

/* Defines */
#define WEB_VIEW_MIN_WIDTH               900
#define WEB_VIEW_MIN_HEIGHT              700
#define URL_RETRY_INTERVAL               2000000
#define WEB_VIEW_RELEASE_POINTER(ptr)    releasePointer((void **)&ptr);
#define WEB_VIEW_SECURE_ZERO_MEMORY(ptr) secureZeroMemory(ptr, ptr == NULL ? 0 : strlen(ptr) * sizeof(char));
#define INSPECTOR FALSE


/* Globals */
extern const char   *gicaroot;    /* Defined in UIDialogGtk.c */
extern unsigned int  option_bits; /* Defined in UIDialogGtk.c */
/* Local Globals */
static pthread_mutex_t  local_mutex       = PTHREAD_MUTEX_INITIALIZER; /* Mutex to synchronize the operations on the wait condition. */
static GtkWidget       *window            = NULL;                      /* The main window of the webview. */
static char            *response          = NULL;                      /* The response returned by the webview. */
static webkit_option    webkit_settings[] = {                          /* The settings to apply to the webview. */
    { "enable-default-context-menu",               FALSE }, /* Since WebKit 1.1.18 */
    { "enable-html5-database",                     FALSE }, /* Since WebKit 1.1.8 */
    { "enable-html5-local-storage",                FALSE }, /* Since WebKit 1.1.8 */
    { "enable-offline-web-application-cache",      TRUE }, /* Since WebKit 1.1.13 */ //Timeout was reached fix: This needs to be enabled to speed load time, else AuthPage timeout may happens
    { "enable-display-of-insecure-content",        FALSE }, /* Since WebKit 2.0 */
    { "enable-dns-prefetching",                    FALSE }, /* Since WebKit 1.1.13 */
    { "enable-java-applet",                        FALSE }, /* Since WebKit 1.1.22 */
    { "enable-plugins",                            FALSE }, /* No "Since" specified so assuming WebKit 1.0 */
    { "enable-running-of-insecure-content",        FALSE }, /* Since WebKit 2.0 */
    { "media-playback-allows-inline",              FALSE }, /* Since WebKit 1.10.0 */
    {  "enable-developer-extras",                  INSPECTOR},
    { NULL,                                        FALSE }
};


namespace UIDialogLibWebKit{


        /* Callback on the "window.external.citrixExitWebview(samlResponse)"
   javascript function in the webview. */
        void onCitrixExitWebview(int args1[], std::string args2[]){

            g_t1.logInfo("UIDIALOGLIB: Entered in UIDialogLibWebKit::onCitrixExitWebview \n");
            //Get the token
            response = (char*)malloc(args2[0].length()*2);
            strcpy(response, args2[0].c_str());

            //Should be called after closing the webextension
            g_idle_add(closeWebview, NULL);

            return;
        }


        PF* getFunctionArray()
        {

            PF *func_array = new PF[TOTAL_FUNCTIONS + 1];
            func_array[enum_onCitrixExitWebview] = UIDialogLibWebKit::onCitrixExitWebview;
            return func_array;

        }
        /*static */ void initWebExtension(WebKitWebContext *context, GVariant* unused)
        {

            g_signal_stop_emission_by_name(context, "initialize-web-extensions");


            GVariant* user_data = g_variant_new("s", gicaroot);
            std::string icaroot = std::string(gicaroot);
            if(icaroot.empty()){
                std::cerr << "Couldn't find icaroot, try setting ICAROOT environment variable and try again, exiting\n";
                exit(-1);
            }

            std::string lib_location_path = icaroot + "/lib/UIDialogLibWebKit3_ext";
            g_t1.logInfo("UIDIALOGLIB: This is the lib_location_path, expecting to find extension lib here: %s\n", lib_location_path.c_str());

            webkit_web_context_set_web_extensions_directory(context, lib_location_path.c_str());
            if(user_data != NULL){
                webkit_web_context_set_web_extensions_initialization_user_data(context, user_data);
            }

           
        }
}



//Timeout was reached fix: Below function is called if timeout was reached signal is fired and
// instead of throwing error page reload occurs .
/* static */ void loadFailed(WebKitWebView  *web_view,
                                        WebKitLoadEvent load_event,
                                        gchar *failing_uri,
                                        GError *error,
                                        callback_data *data)
{
    std::cout<<"Loading of page failed with error :"<<error->message<<"\n";
    std::cout<<"Error throwing uri is: "<<failing_uri<<"\n";
    std::cout<<"Loading page again...\n";
    std::cout<<"Waiting for "<<URL_RETRY_INTERVAL<<" sec before reloading\n";
    g_t1.logInfo("Loading of page failed with error %s. Uri is %s. Waiting for %d secs before reloading.", error->message, failing_uri, (int)URL_RETRY_INTERVAL);
    usleep(URL_RETRY_INTERVAL); /* If we load page too fast the effectiveness of reload will not be there, since traffic may be high during timeout and ease out should happen in interval of reloading*/
    std::cout<<"Reloading uri ..\n";
    g_t1.logInfo("Reloading...");
    webkit_web_view_load_uri(web_view,failing_uri);
    
}

/*static */ void web_view_load_changed (WebKitWebView  *web_view,
                                   WebKitLoadEvent load_event,
                                   callback_data        *user_data)
{
    switch (load_event) {
    case WEBKIT_LOAD_STARTED:
        /*Intimation to user when auth load started*/
        std::cout<<"Load of Auth page started..\n";
        g_t1.logInfo("Load of Auth page started..");
        break;
    case WEBKIT_LOAD_REDIRECTED:
        /*Intimation to user the url is redirected*/
        const gchar* redirected_uri;
        redirected_uri= webkit_web_view_get_uri (web_view);
        std::cout<<"Page is redirected \n";
        g_t1.logInfo("Page is redirected to : %s\n",redirected_uri);
        break;
    case WEBKIT_LOAD_COMMITTED:
        /*Intimation to user that uri loading started*/
        const gchar *uri ;
        uri = webkit_web_view_get_uri (web_view);
        std::cout<<"Loading uri..";
        g_t1.logInfo("Loading uri: %s\n",uri);
        break;
    case WEBKIT_LOAD_FINISHED:
        /*Intimation to user that loading is finished.*/
        std::cout<<"Load finished..\n";
        g_t1.logInfo("Load finished..\n");
        break;
    }
}
/* Implements the UIDialogDisplayWebview() function from the UI Dialog Library SDK. */
const char** UIDialogDisplayWebview(const char *url,
                                    const char *window_title,
                                    const char *post_data,
                                    UIDIALOG   *parent)
{
    // perror("Entered UIDialogDisplayWebview \n");
    g_t1.logInfo("Entered UIDialogDisplayWebview \n");
    webkit_server *ws;
    std::string icaclient, sock_path;
    icaclient = LinuxCommonUtils::getHomeDir();
    icaclient = icaclient + std::string("/.ICAClient");

    if(!LinuxCommonUtils::ifFileExists(icaclient)){
        std::cerr <<  icaclient << " folder not created, anomoly use case, exiting\n";
        exit(0);

    }

    sock_path = icaclient + std::string("/webkit_socket1");


    //Initialize the server part of the communication
    ws = new webkit_server(UIDialogLibWebKit::getFunctionArray(), sock_path);

    webview_data *web_view_data = NULL;

    /* Clear the response if necessary. */
    WEB_VIEW_SECURE_ZERO_MEMORY(response);
    WEB_VIEW_RELEASE_POINTER(response);

    /* Prepare the webview data. */
    web_view_data = (webview_data*) malloc(sizeof(webview_data));

    pthread_cond_t wait_condition;
    pthread_cond_init(&wait_condition, NULL);

    web_view_data->url            =  url;
    web_view_data->window_title   =  window_title;
    web_view_data->post_data      =  post_data;
    web_view_data->parent         =  parent;
    web_view_data->wait_condition = &wait_condition;

    /* Display the web view and hold onto it until the wait condition has been signalled. */
    pthread_mutex_lock(&local_mutex);
    g_idle_add((gboolean(*)(gpointer user_data)) displayWebview, web_view_data);
    pthread_cond_wait(&wait_condition, &local_mutex);
    pthread_cond_destroy(&wait_condition);
    pthread_mutex_unlock(&local_mutex);


    ws->deinitComm();
    free(ws);
    // printf("This is what we are returning to AuthManagerDaemon %s\n", response);
    return (const char**)&response;
}
gboolean onDecidePolicy (WebKitWebView           *web_view,
                        WebKitPolicyDecision    *decision,
                        WebKitPolicyDecisionType decision_type,
                        gpointer                 user_data)
{

    gchar* command;
    switch (decision_type) {

    case WEBKIT_POLICY_DECISION_TYPE_NEW_WINDOW_ACTION:;
        WebKitNavigationPolicyDecision *navigation_decision = WEBKIT_NAVIGATION_POLICY_DECISION (decision);
        if(strcmp(webkit_navigation_policy_decision_get_frame_name(navigation_decision),"_blank")==0)       /* Navigate blank targets to new tab in native browser */
            command = g_strconcat("xdg-open ",webkit_uri_request_get_uri (webkit_navigation_action_get_request(webkit_navigation_policy_decision_get_navigation_action(navigation_decision))),NULL);
            system(command);
    
        break;

    }
    return TRUE;   /* Return true so the event can be propogated to other handlers */
}

/* Creates a GTK window with a web view and displays it. */
/* static */ gboolean displayWebview(webview_data *webviewData)
{

    g_t1.logInfo("UIDIALOGLIB: Entred displaywebview\n");
    
    /* Initialize communication and give the directory path to locate webextension.so with it */
    g_signal_connect(webkit_web_context_get_default(), "initialize-web-extensions", G_CALLBACK(UIDialogLibWebKit::initWebExtension), NULL);

    callback_data *on_destroy_data = NULL;

    /* Create the window. */
    window = gtk_window_new(GTK_WINDOW_TOPLEVEL);

    /* Set the callback for when the window is destroyed. */
    on_destroy_data = (callback_data*)malloc(sizeof(callback_data));
    on_destroy_data->wait_condition = webviewData->wait_condition;
    on_destroy_data->signal         = g_signal_connect(G_OBJECT(window),
                                                       "destroy",
                                                       G_CALLBACK(onDestroy),
                                                       on_destroy_data);
    /*Rewritten

    /* Monitor for the ESC key being pressed. */
    g_signal_connect(G_OBJECT(window),
                     "key_press_event",
                     G_CALLBACK(onKeyPressEvent),
                     NULL);
    /* Set the minimum size. */
    gtk_widget_set_size_request(window, WEB_VIEW_MIN_WIDTH, WEB_VIEW_MIN_HEIGHT);

    /* Set the dialogs parent. */
    if (webviewData->parent != NULL)
    {
        gtk_window_set_modal(GTK_WINDOW(window), TRUE);
        gtk_window_set_transient_for(GTK_WINDOW(window), GTK_WINDOW(webviewData->parent->window));

        /* Set dialog position */
        gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER_ON_PARENT);

        if (option_bits & UI_DIALOGS_ABOVE)
        {
            gtk_window_set_keep_above(GTK_WINDOW(window), TRUE);
        }
    }
    else
    {
        /* Set dialog position */
        gtk_window_set_modal(GTK_WINDOW(window), TRUE);
        gtk_window_set_position(GTK_WINDOW(window), GTK_WIN_POS_CENTER);

        /* We don't have a parent and we want to be seen
           so set the dialog window to on top. */
        gtk_window_set_keep_above(GTK_WINDOW(window), TRUE);
    }

    /* Set the dialog title. */
    gtk_window_set_title(GTK_WINDOW(window), webviewData->window_title);

    /* Set the dialog icon. */
    gchar *icon_path = g_strconcat(gicaroot, "/icons/000_Receiver_64.png", NULL);
    gtk_window_set_default_icon_from_file(icon_path, NULL);
    g_free(icon_path);

    /* Create a WebKitGTK+ view. */
    GtkWidget *web_view = webkit_web_view_new();

    /* Configure the web view. */
    configureWebview(WEBKIT_WEB_VIEW(web_view));
    /* Attach signal to decide on navigation policy of blank targets */
    g_signal_connect(web_view, "decide-policy", G_CALLBACK(onDecidePolicy),NULL);

    /* Put the browser area into the main window. */
    gtk_container_add(GTK_CONTAINER(window), web_view);

    /* Load the web page into the browser instance. */
    g_t1.logInfo("This is the loaded weburl %s", webviewData->url);
    webkit_web_view_load_uri(WEBKIT_WEB_VIEW(web_view), webviewData->url);
   

    g_signal_connect(WEBKIT_WEB_VIEW(web_view), "load-failed",
                                        G_CALLBACK(loadFailed),
                                        NULL);
    
    g_signal_connect(WEBKIT_WEB_VIEW(web_view), "load-changed",
                                        G_CALLBACK(web_view_load_changed),
                                        NULL);

    /* Make sure that when the browser area becomes visible, it will get mouse
       and keyboard events. */
    gtk_widget_grab_focus(web_view);



    /* Display the dialog. */
    gtk_widget_show_all(window);
     
    #if INSPECTOR
         WebKitWebInspector *inspector = webkit_web_view_get_inspector(WEBKIT_WEB_VIEW(web_view));
         webkit_web_inspector_show(inspector);
        
    #endif
    /*gtk_main();  //This is commented since gtk_main() call is invoked in init of uidialoglib and exit of uidialoglib had gtk_main_quit. This can be uncommented when any issues is observed in loading UI.*/

    /* Release the callback data. */
    WEB_VIEW_RELEASE_POINTER(webviewData);

    return FALSE; /* The code we were signalled for has been run so remove from
                     the GTK+ mainloop. */
}

/* Invokes the destruction of the main window. */
/* static */ gboolean closeWebview(void* args)
{
    if (window != NULL)
    {
        gtk_widget_destroy(window);
    }

    return FALSE; /* The code we were signalled for has been run so remove from
                     the GTK+ mainloop. */
}

/* Configures the web view parameters. */
/* static */ void configureWebview(WebKitWebView *webview)
{
    /* Get WebKit Settings. */
    GObject *settings = G_OBJECT(webkit_web_view_get_settings(webview));
    GObjectClass *settings_class = G_OBJECT_GET_CLASS(settings);

    /* Loop through the webkit options to set. */
    webkit_option *webkit_option_ptr = webkit_settings;
    while (webkit_option_ptr->name != NULL)
    {
        /* Set the webkit option. */
        GParamSpec *param_spec = g_object_class_find_property(settings_class,
                                                              webkit_option_ptr->name);
        if (param_spec != NULL)
        {
            g_object_set(settings, webkit_option_ptr->name, webkit_option_ptr->value, NULL);
        }

        /* Move to the next option. */
        webkit_option_ptr++;
    }


    GParamSpec *parm = g_object_class_find_property(G_OBJECT_GET_CLASS(G_OBJECT(settings)), "hardware-acceleration-policy");
    if(parm != NULL)
            g_object_set(G_OBJECT(settings), "hardware-acceleration-policy", WEBKIT_HARDWARE_ACCELERATION_POLICY_NEVER, NULL);


}

/* Callback on clicking a key when the webview window has focus. */
/* static */ gboolean onKeyPressEvent(GtkWidget   *unused_widget,
                                      GdkEventKey *event)
{
    /* Annoyingly the Escape key has had its name changed. GDK_Escape is the
       older of the two defines. */
#ifdef GDK_KEY_Escape
    if (event->keyval == GDK_KEY_Escape)
#else
    if (event->keyval == GDK_Escape)
#endif
    {
        /* The Escape key has been pressed. */

        /* The default cancel response is NULL. */
        WEB_VIEW_SECURE_ZERO_MEMORY(response);
        WEB_VIEW_RELEASE_POINTER(response);

        /* The webview can be closed, so we
           just register a close event in the GTK+ event loop.
           This will trigger the "destroy" event and call the "on_destroy" callback,
           where the wait condition is signalled to the UI thread.*/
        g_idle_add( closeWebview, NULL);

        return TRUE; /* Stop other handlers from being invoked for this event. */
    }

    return  FALSE; /* Propagate the event further as normal. */
}




/* Callback to signal that the window is being destroyed. */
/* static */ void onDestroy(GtkWidget     *unused_widget,
                            callback_data *onDestroyData)
{
    /* Disconnect the calling signal to avoid recursion. */
    if (window != NULL)
    {
        g_signal_handler_disconnect(window, onDestroyData->signal);
    }

    /* Signal that the window is closed. */
    pthread_mutex_lock(&local_mutex);
    pthread_cond_signal(onDestroyData->wait_condition);
    pthread_mutex_unlock(&local_mutex);

    /* Release the callback data. */
    WEB_VIEW_RELEASE_POINTER(onDestroyData);
}

gboolean sUIDialogClearWebviewCookies(void* unused){
    WebKitCookieManager* cookie_manager = webkit_web_context_get_cookie_manager (webkit_web_context_get_default ());
    webkit_cookie_manager_delete_all_cookies(cookie_manager);
    return FALSE;
}

/*
Implements the UIDialogClearWebviewCookies() function from the UI Dialog Library SDK. 
Currently we are using a deprecated function, there is no synchronous function alternative in the newer webkit versions. 
In future we might have to find a way to force a async function to sync function.
*/
void UIDialogClearWebviewCookies(void)
{
    g_idle_add(sUIDialogClearWebviewCookies, NULL);
}

